!(c) 2012 Hübbe, Körner
!
! This program is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mo_utilities
	use mpi
	implicit none
contains

	subroutine initMatrix(matrix)
		double precision, dimension(:,:), intent(out) :: matrix
		integer :: processCount, trash, curProcess, height, y0, y1

		call mpi_comm_size(mpi_comm_world, processCount, trash)
		height = ubound(matrix, 2)
		do curProcess = 0, processCount-1
			y0 = curProcess*height/processCount + 1
			y1 = (curProcess + 1)*height/processCount
			matrix(:,y0:y1) = curProcess
		end do
	end subroutine

	function intToStr(a)
		character(len=11) :: intToStr
		integer, intent(in) :: a
		write(intToStr, '(i11)') a
		intToStr = trim(adjustl(intToStr))
	end function

	subroutine waste_cycles(iterations)
		integer, intent(in) :: iterations
		integer, save :: trash
		integer :: i

		do i = 1, iterations
			trash = trash + i
		end do
	end subroutine

	subroutine dumpMatrix(matrix)
		double precision, dimension(:,:), intent(in) :: matrix
		integer :: width
		character(len=22) :: formatString

		width = ubound(matrix, 1)
		formatString = '(' // intToStr(width) // 'i2)'
		write(*, formatString) int(matrix)
		write(*,*)
		call waste_cycles(1000*1000)	!Actually, the thing to do here would be to flush the output. But I don't know how to do that in Fortran, and it's running late.
	end subroutine

	subroutine updateJacobi(inputMatrix, outputMatrix)
		double precision, dimension(:,:), intent(in) :: inputMatrix
		double precision, dimension(:,:), intent(out) :: outputMatrix
		integer :: width, height, myProcessId, processCount, curProcess, trash

		!Output the input matrices in process order.
		call mpi_comm_rank(mpi_comm_world, myProcessId, trash)
		call mpi_comm_size(mpi_comm_world, processCount, trash)
		do curProcess = 0, processCount-1
			call mpi_barrier(mpi_comm_world, trash)	!Yes, here we need it so that the output is nicely separated.
			if (curProcess == myProcessId) call dumpMatrix(inputMatrix)
			call mpi_barrier(mpi_comm_world, trash)	!Yes, here we need it so that the output is nicely separated.
		end do

		!Do a faked update of the outputMatrix.
		width = ubound(inputMatrix, 1)
		height = ubound(outputMatrix, 2)
		outputMatrix(2:width-1,2:height-1) = inputMatrix(2:width-1,2:height-1) + 1
		outputMatrix(:,1) = inputMatrix(:,1)
		outputMatrix(:,height) = inputMatrix(:,height)
		outputMatrix(1,:) = inputMatrix(1,:)
		outputMatrix(width,:) = inputMatrix(width,:)
	end subroutine

end module

program dummy
	use mo_utilities
	implicit none

	double precision :: matrixA(30,5), matrixB(30,5)
	integer :: myProcessId, trash

	call mpi_init(trash)

	call initMatrix(matrixA)
	call updateJacobi(matrixA, matrixB)
	call mpi_barrier(mpi_comm_world, trash)	!Yes, here we need it so that the output is nicely separated.
	call mpi_comm_rank(mpi_comm_world, myProcessId, trash)
	if (myProcessId == 0) call dumpMatrix(matrixB)

	call mpi_finalize(trash)

end program dummy
